


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"
#include "interface.h"
#include "callbacks.h"
#include "dlgbox.h"
#include "support.h"
#include "functions.h"
#include "application.h"
#include "dso.h"
#include "macro.h"
#include "miscop.h"


/*!
\defgroup mainlevel Using Crystal Image

\brief Command-line arguments

<ul>
<li>	-h -help          print command-line usage
<li>	-l -load          Load an image file
<li>	-m -macro         Load a macro file
<li>	-q -quiet         Disable debug messages (debug level 0)
<li>	-v -verbose       Enable debug messages at debug level 2
</ul>
<p>

\brief Configuration file

The configuration file is stored in the user's home directory
under the name .cimagerc. The file will be created during the 
first run if it does not exist. Contents of the file can be edited
using the <i>File->Settings</i> menu.

*/

int read_prefs ()
{
FILE* FP;
char fname [256];
char s[256];
char s1[10];
char* home;
int line,i,j;

	/* Do this silently because gtk is not set up at this point */
	
	home = getenv ("HOME");
	if (!home) return -1;

	strcpy (fname, home);
	strcat (fname, "/.cimagerc");
	FP = fopen (fname, "rt");
	if (!FP)
	{
		strcpy (fname, "/etc/cimage.conf");
		FP = fopen (fname, "rt");
	}
	if (!FP) return -1;			/* no preferences file. Too bad. */

	/* Read the file line by line */

	line=0;
	while (fgets (s, 255, FP) != NULL)
	{
		line++;
		for (i=0; i<9; i++) s1[i]=0;
		if ((!isempty(s)) && (s[0]!='#'))	/* not a comment line */
		{
			if (!strncasecmp (s, "tiffinfo", 8)) 
			{
				i=8;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				tiffreport = atoi (s1);
				if ((tiffreport!=0) && (tiffreport!=1))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", tiffreport,line);
					tiffreport=0;
				}
			}
			else if (!strncasecmp (s, "showrulers", 10)) 
			{
				i=10;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				show_rulers = atoi (s1);
				if ((show_rulers!=0) && (show_rulers!=1))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", show_rulers,line);
					show_rulers=0;
				}
			}
			else if (!strncasecmp (s, "interpolation", 13)) 
			{
				i=13;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				intp_pref = atoi (s1);
				if ((intp_pref<0) || (intp_pref>6))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", intp_pref,line);
					intp_pref=1;
				}
			}
			else if (!strncasecmp (s, "expand_contrast", 15)) 
			{
				i=15;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				autoscale = atoi (s1);
				if ((autoscale<0) || (autoscale>1))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", intp_pref,line);
					autoscale=1;
				}
			}
			else if (!strncasecmp (s, "debuglevel", 10)) 
			{
				i=10;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				debuglevel = atoi (s1);
				if ((debuglevel<0) || (debuglevel>3))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", debuglevel,line);
					debuglevel=0;
				}
			}
			else if (!strncasecmp (s, "skip_license", 12)) 
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				skip_license = atoi (s1);
				if ((debuglevel<0) || (debuglevel>3))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", skip_license,line);
					skip_license=0;
				}
			}
			else if (!strncasecmp (s, "undo_function", 13)) 
			{
				i=13;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,13));
				undo_function = atoi (s1);
				if ((undo_function<0) || (undo_function>2))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", debuglevel,line);
					undo_function=0;
				}
			}
			else if (!strncasecmp (s, "path_to_browser", 15)) 
			{
				i=15;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (browspath, 0, 1020);
				strncpy (browspath, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "path_to_gnuplot", 15)) 
			{
				i=15;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (gnuplotpath, 0, 1020);
				strncpy (gnuplotpath, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "path_to_help", 12)) 
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (helppath, 0, 1020);
				strncpy (helppath, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "path_to_dsos", 12)) 
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (dsopath, 0, 1020);
				strncpy (dsopath, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "path_to_undo", 12)) 
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (undo_path, 0, 1020);
				strncpy (undo_path, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "sv_viewer_name", 14)) 
			{
				i=14;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (imager_sv_path, 0, 1020);
				strncpy (imager_sv_path, &s[i], MIN(1000,j-i));
			}
			else
				dp (0,"Warning: Illegal keyword in .cimagerc, line %d\n",line);
		}
	}

	fclose (FP);
	return 0;
}



/********************************************************************/


#define nextargi (--argc,atoi(*++argv))
#define nextargf (--argc,atof(*++argv))
#define nextargs (--argc,*++argv)

void prhelp (char *title)
{
char buf[256];
char helpmsg[1024];


	helpmsg[0] = 0;
	sprintf (buf, "Crystal Image: A powerful quantitative image analysis tool\n");
	strncat (helpmsg, buf, 1020);
	sprintf (buf, "(c) 2002-2010 Mark A. Haidekker\n");
	strncat (helpmsg, buf, 1020);
	sprintf (buf, "Invalid command-line option. Valid options are:\n");
	strncat (helpmsg, buf, 1020);
	sprintf (buf, "-h --help          print this help text\n");
	strncat (helpmsg, buf, 1020);
	sprintf (buf, "-l --load          Load an image file\n");
	strncat (helpmsg, buf, 1020);
	sprintf (buf, "-m --macro         Load a macro file\n");
	strncat (helpmsg, buf, 1020);
	sprintf (buf, "-q --quiet         Disable debug messages\n");
	strncat (helpmsg, buf, 1020);
	sprintf (buf, "-v --verbose       Enable debug messages (repeated -v increases verbosity)\n");
	strncat (helpmsg, buf, 1020);
	
	if (title)
		ShowLongMessage (helpmsg, title);
	else
		ShowLongMessage (helpmsg, "Crystal Image command-line help");

}



int cmdline_args (int argc,char **argv)
{
char fname[1024];


    while (--argc && **++argv=='-') 
	{

		if (!strcmp(argv[0],"-l") || !strcmp(argv[0],"--load"))
		{
			if (argv[1])
			{
				strcpy (fname,nextargs);
				dp (1, "Command-line initiated load of file %s\n",fname);
				file_read (fname);
			}
			
		}
		else if (!strcmp(argv[0],"-q") || !strcmp(argv[0],"--quiet"))		/* Override debuglevel and set to 0 */
			{ debuglevel=0; }

		else if (!strcmp(argv[0],"-v") || !strcmp(argv[0],"--verbose"))		/* Override debuglevel and set to 2 */
			{ debuglevel++; }
		
		else if (!strcmp(argv[0],"-m") || !strcmp(argv[0],"--macro"))
		{
			if (argv[1])
			{
				strcpy (fname,nextargs);
				dp (1, "Command-line initiated load of macro file %s\n",fname);
				load_macro (fname);
			}
			
		}
        else 
		{
        	if (strcmp(argv[0],"--help") && strcmp (argv[0],"-h"))
        		prhelp ("Error:Unknown command-line option");
        	else
				prhelp (NULL);
		}
	}
	return 0;
}



/*************************************

	#     #
	##   ##    ##       #    #    #
	# # # #   #  #      #    ##   #
	#  #  #  #    #     #    # #  #
	#     #  ######     #    #  # #
	#     #  #    #     #    #   ##
	#     #  #    #     #    #    #

***************************************/


int main (int argc, char *argv[])
{
 GtkWidget *window1;	/* Main (menu) window */
 GdkEvent *event;
 int nevents;
 

 	gtk_init (&argc, &argv);

	add_pixmap_directory (PACKAGE_DATA_DIR "/" PACKAGE "/pixmaps");


	/* Initialize the Crystal Image application...*/
	
	app_init ();

	window1 = create_window1 ();
	gtk_widget_show (window1);		/* Display the control window */


	/* Read .rc file. If nonexistent, create one */
	
	if (read_prefs () ==-1)
	{
		display_license_splash ();
		save_default_prefs ();
		read_prefs ();
	}
	
	/* Load the DSOs. */
	
	load_dso ();

	/* Get the path with getcwd to populate imager_cwd  */

	imager_cwd = calloc (2048, sizeof (char));
	getcwd (imager_cwd, 2040);
	dp (2,"Current working directory set to %s\n",imager_cwd);


	
    /* Get gdk to use imlib's visual and colormap */
    ///gtk_widget_push_visual(gdk_imlib_get_visual());
    ///gtk_widget_push_colormap(gdk_imlib_get_colormap());
	gtk_widget_push_visual   (gdk_rgb_get_visual());
	gtk_widget_push_colormap (gdk_rgb_get_cmap());

  	gtk_widget_show (window1);		/* Display the control window */

	/* NOTE: By removing or altering the next line you indicate acceptance of the terms and conditions of the license */
	if (!skip_license) display_license_splash ();		/* Display a splash screen for 5 seconds with the license */


	/* At this time, GTK is up and the main window has been created. 
		handle possible command-line options *BEFORE* we run into
		GTK's main event loop.
		At the ssme time, we also define the default icon 
		for all windows and sub-windows. */

	cmdline_args (argc, argv);
	gtk_window_set_default_icon (create_icon());


  	gtk_main ();
  	return 0;
}

