

/*
	wvlt_core.c : Core functions for wavelet transforms
	Please note that this module is somewhat experimental
	and may still contain hidden bugs.

	Credits: The core functions of the DWT, particularly the 
	subband filters, are leaning heavily on Numerical Recpies in C,
	second edition, chapter 13.

 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <string.h>
#include <math.h>

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "application.h"
#include "dlgbox.h"
#include "filter.h"
#include "functions.h"
#include "functions2.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "secondwnd.h"
#include "wvlt_core.h"






/***************************************************************************************************************

  ***    *   *   ****    ****      *     *   *   ****            *****   *       *****
 *   *   *   *   *   *   *   *    * *    *   *   *   *           *       *         *
 *       *   *   *   *   *   *   *   *   **  *   *   *           *       *         *
  ***    *   *   ****    ****    *   *   * * *   *   *           ****    *         *
     *   *   *   *   *   *   *   *****   *  **   *   *           *       *         *
 *   *   *   *   *   *   *   *   *   *   *   *   *   *           *       *         *
  ***     ***    ****    ****    *   *   *   *   ****            *       *****     *

*/



/* The subband filters perform one step of DWT decomposision or synthesis.
	These subband filters operate on an 1D array of float data (length
	must be a power of two or - less strictly - an even number) and return
	the scaling filter coefficients in the lower half of the array and the 
	wavelet coefficients in the upper half of the array.
	The variable idwt determines whether we want a forward DWT (idwt==0)
	or an inverse DWT (idwt!=0).
*/



void dead_subband (float *data, unsigned long n, DWTDIRECTION idwt, wavefilt dummy)
{
	dp (0, "Dead subband filter. Should not happen.\n");
}



/*
	Initializing routine for general_subband. Here, the wavefilt structure is initialized and
	filled with the wavelet and scale parameters.
	This function MUST be called before general_subband can be used.
	Note that the standalone functions in the wavelet *module* make use of specific 
	parameter set lengths and are therefore faster than general_subband.
*/


/* w00t! Check out http://wavelets.pybytes.com/ */


wavefilt wavelet_prepare (WAVELET_TYPE wlt)
{
int k;
float sgn;
wavefilt wfilt;

	/* scale filter (lowpass) parameters h(k). */

static float c_haar[2] = 	{	0.7071067118655, 0.7071067118655 };					/* Haar */

static float c_daub4[4] =	{	0.4829629131445341,0.8365163037378079,				/* DAUB4, 2 vanishing moments */
								0.2241438680420134,-0.1294095225512604 };

static float c_daub8[8] = 	{	 0.23037781330890,  0.71484657055292, 				/* DAUB8, 4 vanishing moments */
								 0.63088076792986, -0.02798376941686,
								-0.18703481171909,  0.03084138183556,
								 0.03288301166689, -0.01059740178507 };

static float c_daub12[12]=	{	0.111540743350, 0.494623890398, 0.751133908021,		/* DAUB12, 6 vanishing moments */
								0.315250351709,-0.226264693965,-0.129766867567,
								0.097501605587, 0.027522865530,-0.031582039318,
								0.000553842201, 0.004777257511,-0.001077301085 };

static float c_daub20[20] =	{	0.026670057901, 0.188176800078, 0.527201188932,		/* DAUB20, 10 vanishing moments */
								0.688459039454, 0.281172343661,-0.249846424327,
								-0.195946274377, 0.127369340336, 0.093057364604,
								-0.071394147166,-0.029457536822, 0.033212674059,
								0.003606553567,-0.010733175483, 0.001395351747,
								0.001992405295,-0.000685856695,-0.000116466855,
								0.000093588670,-0.000013264203 };

static float c_coif6[6] = 	{	-0.07273261951285,  0.33789766245781,				/* COIF6, 2 vanishing moments */
								 0.85257202021226,  0.38486484686420,
								-0.07273261951285, -0.01565572813546 };

static float c_coif12[12] =	{	 0.016387336463, -0.041464936781,					/* COIF12, 4 vanishing moments */
								-0.067372554722,  0.386110066823,
								 0.812723635449,  0.417005184423,
								-0.076488599078, -0.059434418646,
								 0.023680171946,  0.005611434819,
								-0.001823208870, -0.000720549446 };

static float c_symlet4[8] =	{	0.032223100604042702, -0.012603967262037833,
									-0.099219543576847216, 0.29785779560527736,
									0.80373875180591614, 0.49761866763201545,
									-0.02963552764599851, -0.075765714789273325};


static float c_symlet8[16] =	{	0.0018899503327594609, -0.0003029205147213668,	/* Symlet */
									-0.014952258337048231, 0.0038087520138906151,
									0.049137179673607506, -0.027219029917056003,
									-0.051945838107709037, 0.3644418948353314,
									0.77718575170052351, 0.48135965125837221, 
									-0.061273359067658524, -0.14329423835080971,
									0.0076074873249176054, 0.031695087811492981,
									-0.00054213233179114812, -0.0033824159510061256 };

static float c_symlet12[24] =	{	-0.00017906658697508691, -1.8158078862617515e-05, 
									0.0023502976141834648, 0.00030764779631059454,
									-0.014589836449234145, -0.0026043910313322326,
									0.057804179445505657, 0.01530174062247884,
									-0.17037069723886492, -0.07833262231634322,
									0.46274103121927235, 0.76347909778365719,
									0.39888597239022, -0.022162306170337816,
									-0.035848830736954392, 0.049179318299660837,
									0.0075537806116804775, -0.024220722675013445,
									-0.0014089092443297553, 0.007414965517654251,
									0.00018021409008538188, -0.0013497557555715387,
									-1.1353928041541452e-05, 0.00011196719424656033};



	wfilt.n_param=0;
	
	if (wlt==W_HAAR) 
	{
		wfilt.n_param=2;
		wfilt.hk=c_haar;
	}
	else if (wlt==W_DAUB4) 
	{
		wfilt.n_param=4;
		wfilt.hk=c_daub4;
	}
	else if (wlt==W_DAUB8) 
	{
		wfilt.n_param=8;
		wfilt.hk=c_daub8;
	}
	else if (wlt==W_DAUB12) 
	{
		wfilt.n_param=12;
		wfilt.hk=c_daub12;
	}
	else if (wlt==W_DAUB20) 
	{
		wfilt.n_param=20;
		wfilt.hk=c_daub20;
	}
	else if (wlt==W_COIF6) 
	{
		wfilt.n_param=6;
		wfilt.hk=c_coif6;
	}
	else if (wlt==W_COIF12) 
	{
		wfilt.n_param=12;
		wfilt.hk=c_coif12;
	}
	else if (wlt==W_SYM4) 
	{
		wfilt.n_param=8;
		wfilt.hk=c_symlet4;
	}
	else if (wlt==W_SYM8) 
	{
		wfilt.n_param=16;
		wfilt.hk=c_symlet8;
	}
	else if (wlt==W_SYM12) 
	{
		wfilt.n_param=24;
		wfilt.hk=c_symlet12;
	}
	else 
		ShowMessage ("Error setting up wavelet parameters");

	wfilt.gk = (float*) calloc (wfilt.n_param, sizeof (float));

	/* Generate the wavelet filter parameters g(k) from the scale filter h(k) */
	
	sgn = -1.0;
	for (k=0; k<wfilt.n_param; k++) 
	{
		wfilt.gk[wfilt.n_param-1-k]=sgn*wfilt.hk[k];
		sgn = -sgn;
	}
	
	dp (2, "Setting up filter parameters: You have:\n");
	dp (2, "Lowpass:  ");
	for (k=0; k<wfilt.n_param; k++) dp (2, "%7.4f ",wfilt.hk[k]); dp (2,"\n");
	dp (2, "Highpass: ");
	for (k=0; k<wfilt.n_param; k++) dp (2, "%7.4f ",wfilt.gk[k]); dp (2,"\n");
	

/*	
	These values center the "support" of the wavelets at each level. Alternatively, the "peaks"
	of the wavelets can be approximately centered by the choices hk_offs=-2 and gk_offs=-n+2.
	Note that daub4 and this function with n=4 use different default centerings.
*/
	wfilt.hk_offs = wfilt.gk_offs =  0; //-(n >> 1);
	return wfilt;

}



/*
	Generalized one-stage subband filter: applies an arbitrary wavelet filter to data  (for idwt==0)
	or applies its transpose (for idwt==1). Used hierarchically by the actual DWT functions.
	The actual filter is determined by a preceding (a required) call to wavelet_prepare, which initializes
	the structure wfilt.
*/


void general_subband (float *data, unsigned long n, DWTDIRECTION idwt, wavefilt wfilt)
{
float *buf;
unsigned long i,j,k,bitmask,ni,nj,nh,nmod;

	if (n < 4) return;
	
	buf= (float *) calloc (n+1, sizeof (float));		/* temporary storage buffer */

	nmod=wfilt.n_param*n; 						/* A positive constant equal to zero mod n. */
	bitmask=n-1;									/* Mask of all bits, since n is a power of 2. */
	nh = n >> 1;									/* Start of highpass data */

	if (idwt == W_FORWARD)			/* Forward subband coding filter (forward one-level DWT). */
	{
		j=0;
		for (i=0; i<n; i+=2) 					/* Coefficient loop */
		{
			ni = i+nmod+wfilt.hk_offs;				/* Index into data to be incremented and wrapped-around. */
			nj = i+nmod+wfilt.gk_offs;				/* essentially i + a large multiple of n + offset */

			for (k=0; k<wfilt.n_param; k++)			/* Convolution loop */
			{
				buf[j]    += wfilt.hk[k]*data[(ni+k) & bitmask];	/* Scale filter (lowpass) data */
				buf[j+nh] += wfilt.gk[k]*data[(nj+k) & bitmask];	/* Wavelet filter (highpass) data */
			}
			j++;
		}
	} 
	else 			/* Inverse filter (inverse DWT). */
	{
		j=0;
		for (i=0; i<n; i+=2) 
		{
			ni = i+nmod+wfilt.hk_offs;
			nj = i+nmod+wfilt.gk_offs;
			for (k=0; k<wfilt.n_param; k++) 
			{
				buf[(ni+k) & bitmask] += wfilt.hk[k]*data[j];
				buf[(nj+k) & bitmask] += wfilt.gk[k]*data[j+nh];
			}
			j++;
		}
	}
	
	for (i=0; i<n; i++) data[i]=buf[i];        /* Copy the results back from workspace.*/
	free (buf);
}



/* This function initializes the subband() function pointer.
*/


WAVELET_TYPE set_subband (int wavelet)
{
WAVELET_TYPE wtype;
	
	wtype = W_NULL;
	switch (wavelet)
	{
		case 0:						/* Haar */
				wtype=W_HAAR;
				subband = &general_subband;
				strcpy (wvlt_used, "Haar");
				break;
		case 1:						/* DAUB4 */
				wtype=W_DAUB4;
				subband = &general_subband;
				strcpy (wvlt_used, "Daub4");
				break;
		case 2:						/* DAUB8 */
				wtype=W_DAUB8;
				subband = &general_subband;
				strcpy (wvlt_used, "Daub8");
				break;
		case 3:						/* DAUB12 */
				wtype=W_DAUB12;
				subband = &general_subband;
				strcpy (wvlt_used, "Daub12");
				break;
		case 4:						/* DAUB20 */
				wtype=W_DAUB20;
				subband = &general_subband;
				strcpy (wvlt_used, "Daub20");
				break;
		case 5:						/* Coiflet-6 */
				wtype=W_COIF6;
				subband = &general_subband;
				strcpy (wvlt_used, "Coif6");
				break;
		case 6:						/* Coiflet-12 */
				wtype=W_COIF12;
				subband = &general_subband;
				strcpy (wvlt_used, "Coif12");
				break;
		case 7:						/* Symlet 4 */
				wtype=W_SYM4;
				subband = &general_subband;
				strcpy (wvlt_used, "Symlet4");
				break;
		case 8:						/* Symlet 8 */
				wtype=W_SYM8;
				subband = &general_subband;
				strcpy (wvlt_used, "Symlet8");
				break;
		case 9:						/* Symlet12 */
				wtype=W_SYM12;
				subband = &general_subband;
				strcpy (wvlt_used, "Symlet12");
				break;

		default:
				subband = &dead_subband;
				return 0;
	}
	return wtype;
}



/*************************************************************************

 ####    #   #   #####
 #   #   #   #     #
 #   #   #   #     #
 #   #   # # #     #
 #   #   # # #     #
 #   #   # # #     #
 ####     # #      #

*************************************/


/*
	Here are the pyramidal deconstruction/reconstruction algorithms
	for 1D data, n-dimensional data etc. */


/*
	One-dimensional discrete wavelet transform. This routine implements the pyramid algorithm,
	replacing data[0..n-1] by its wavelet transform (for idwt=0), or performing the inverse operation
	(for idwt=1). Note that n MUST be an integer power of 2. The routine wt_subband, whose
	actual name must be supplied in calling this routine, is the underlying wavelet filter. Examples
	of wt_subband are daub4 and (preceded by wavelet_prepare) general_subband.
*/

void dwt_1d (float *data, unsigned long n, DWTDIRECTION idwt, wavefilt wfilt, void (*wt_subband)(float [], unsigned long, DWTDIRECTION, wavefilt))
{
unsigned long i;


	if (n < 4) return;

	if (!idwt)						/* Forward Wavelet transform. */
	{
		for (i=n; i>=4; i>>=1) 
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
		
	} 
	else 							/* Inverse wavelet transform.*/
	{
		for (i=4; i<=n; i<<=1) 
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
	}
}


/* The same as above, but with a limited number of levels */

void dwt_1dn (float *data, unsigned long n, DWTDIRECTION idwt, int levels, wavefilt wfilt, void (*wt_subband)(float *, unsigned long, DWTDIRECTION, wavefilt))
{
unsigned long i;
int l;


	if (n < 4) return;

	if (!idwt)						/* Forward Wavelet transform. */
	{
		i=n; l=levels;
		while ((i>=4) && (l>1))
		{
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
			i >>= 1;
			l--;
		}
		
	} 
	else 							/* Inverse wavelet transform.*/
	{
		i = n; l=levels;
		while (l>1) { i>>=1; l--; }
		l = levels;
		while ((i<=n) && (l>1))
		{
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
			i <<= 1;
			l--;
		}
	}
}



/* Simplistc slice-by-slice 2D wavelet transform. This function
	transforms the current image and returns a float image. */


void dwt_2d (image_type *img, DWTDIRECTION idwt, int decomp_level, wavefilt wfilt, void (*wt_subband)(float *, unsigned long, DWTDIRECTION, wavefilt))
{
image_type result;
float *buf1, *buf2, *p;
int nx, ny;
int x,y,z,i, level;


	/* Expand to nearest power of two */
	
	nx = 1; while (nx<img->xmax) nx=nx<<1;
	ny = 1; while (ny<img->ymax) ny=ny<<1;
	if ((nx<4) || (ny<4))
	{
		ShowMessage ("Not enough data for DWT");
		return;
	}

	/* Get memory */
	
	if (allocate_image (&result, B_FLOAT,nx, ny, img->zmax) ==-1)
		fatal_memerr();
	p = result.data;
	buf1 = calloc (nx, sizeof(float));
	buf2 = calloc (ny, sizeof(float));

	if (idwt==W_INVERSE) 
		dp (1, "2D inverse wavelet transform, %d level(s)...\n", decomp_level);
	else
		dp (1, "2D wavelet transform, %d level(s)...\n", decomp_level);


	for (z=0; z<img->zmax; z++)
	{

		/* Full pyramidal transform in X direction */

		for (y=0; y<img->ymax; y++)
		{
			for (x=0; x<nx; x++) buf1[x]=0;
			for (x=0; x<img->xmax; x++) buf1[x] = readbuf_flt (img, x,y,z);
		
			if (idwt==W_FORWARD)				/* Forward Wavelet transform. */
			{
				i = nx; level=decomp_level;
				while ((i>=4) && (level>0))
				{
					dp (3, "2D DWT in X: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf1, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
					i = i>>1;							/* Divide subband active range by two */
					level--;
				}
				
			} 
			else 							/* Inverse wavelet transform.*/
			{
				i=nx; level = decomp_level;
				while (level>1) { i>>=1; level--; }
				level = decomp_level;
				while ((i<=nx) && (level>0))
				{
					dp (3, "2D IDWT in X: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf1, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
					i = i<<1;
					level--;
				}
			}
	
			for (x=0; x<nx; x++) p[x + nx*(y+ny*z)] = buf1[x];
		}
	
		/* Full pyramidal transform in Y direction */
	
		for (x=0; x<nx; x++)
		{
			for (y=0; y<ny; y++) buf2[y]=p[x + nx*(y+ny*z)];
		
			if (idwt==W_FORWARD)						/* Forward Wavelet transform. */
			{
				i = ny; level=decomp_level;
				while ((i>=4) && (level>0))
				{
					dp (3, "2D DWT in Y: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf2, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
					i = i>>1;							/* Divide subband active range by two */
					level--;
				}
				
			} 
			else 							/* Inverse wavelet transform.*/
			{
				i=ny; level = decomp_level;
				while (level>1) { i>>=1; level--; }
				level = decomp_level;
				while ((i<=ny) && (level>0))
				{
					dp (3, "2D IDWT in Y: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf2, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
					i = i<<1;
					level--;
				}
			}
	
			for (y=0; y<ny; y++) p[x + nx*(y+ny*z)] = buf2[y];
		}
	}

	make_minmax (&result);
	handback (img, &result);

}








/***************************************************************************************************************

 *   *   *   *   *       *****            ***    *   *   ****     ***    *   *   *   *    ***    *   *    ***
 *   *   *   *   *         *             *   *   *   *   *   *     *     *   *   *  *      *     *   *   *   *
 *   *   *   *   *         *             *       *   *   *   *     *     **  *   * *       *     **  *   *
 * * *    * *    *         *              ***    *****   ****      *     * * *   **        *     * * *   *  **
 * * *    * *    *         *                 *   *   *   * *       *     *  **   * *       *     *  **   *   *
 * * *     *     *         *             *   *   *   *   *  *      *     *   *   *  *      *     *   *   *   *
  * *      *     *****     *              ***    *   *   *   *    ***    *   *   *   *    ***    *   *    ***
 
*/



/* Shrinking and thresholding function. Acts on element elem with threshold t.
	Choice is hardsoft=0: Soft thresholding (=shrinking in the strict sense)
				hardsoft=1: hard thresholding
	Returns the shrunk/thresholded element
*/


float wv_thresh (float elem, float t, int hardsoft)
{
	if (!hardsoft)			/* ==0 -> soft thresholding */
	{
		if (fabs(elem) <= fabs (t))
			return 0.0;
		else
			return SIGN(elem)*(fabs(elem)-fabs(t));
	}
	else					/* hard thresholding */
	{
		if (fabs(elem) > fabs (t))
			return elem;
		else
			return 0.0;
	}
}



#define C06745		1.48258			/* this is 1/0.6745 */


/* Estimate Sigma according to Donoho from the highpass coefficients
	of the first decomposition level (i.e., upper half of the signal).
	In its 2D version, we use the H/H square *only* . Therefore, we don't need
	any x,y delimiters, because the H/H area goes from xmax/2 to xmax and
	from ymax/2 to ymax.
*/

float estimate_sigma_1d (float *data, int len)
{
int i,j;
float *p;
float w;


	p = calloc (len/2+10, sizeof (float));
	j=0;
	for (i=len/2; i<len; i++)					/* Find minimum and maximum absolute value in data stream */
	{
		p[j++] = fabs(data[i]);
	}

	w = element_finder (p, j/2, j);	/* Note that this implies an odd n (which we don't have), 
												but with large n the difference is minimal */
	free (p);
	return w*C06745;						/* Donohan & Johnstone estimated standard deviation */
}


float estimate_sigma_2d (image_type *img, int z)
{
int i,n,x,y,xm,ym,xn,yn;
float *p;
float w;

	xn = img->xmax; xm = xn/2;
	yn = img->ymax; ym = yn/2;

	n = (xn-xm+1)*(yn-ym+1);						/* Number of elements, generously over-estimated */
	p = calloc (n, sizeof (float));

	i=0;
	for (y=ym; y<yn; y++)				/* H/H area of level 1 only */
		for (x=xm; x<xn; x++)
		{
			p[i++] = fabs(readbuf_flt (img,x,y,z));
		}

	w = element_finder (p, i/2, i);		/* Note that this implies an odd n (which we don't have), */
	free (p);
	return w*C06745;						/* Donohan & Johnstone estimated standard deviation */
}





/* Wavelet shrinking threshold finder. We pass a 1D array of wavelet coefficients
	(length N) as well as a noise estimate sigma. Sigma can be provided from
	a priori knowledge, or the above Donoho median coefficient estimator can
	be used.

	This function uses the entire 1D dataset, so any calling function needs
	to copy the relevant subband and pass it (or use a suitable pointer).
	Different thresholds can be found for different decomposition levels.
	wshrinkfunc determines the threshold algorithm:
		0: Donoho universal threshold 
		1: BayesShrink, with universal threshold as upper limit
		2: SUREShrink threshold
		3: GCV minimizing threshold

	NOTE that the input field, data, must not be destroyed!

*/


float find_wsthreshold (float *data, int len, float sigma, int wshrinkfunc)
{
int  i,j,cnt;
float T, w, wsigma,min,max, buf, buf2;
float gcv, gcvmin, gcvmint, sumxx, risk, minrisk;
float *p;



	switch (wshrinkfunc)
	{
		case 0:							/* Universal threshold \sigma x \sqrt{2 log n}*/
			T = sigma * sqrt (2*SAFELOG2(len));
			dp (2, "      %d data points yield 'universal' threshold %f\n",len,T);
			return T;

		case 1:							/* Bayes shrink (level-adaptive) */
			
			wsigma = 0; w=0;
			T = sigma * sqrt (2*SAFELOG2(len));		/* Default to Donoho's universal threshold */
			dp (2, "      %d data points yield 'universal' threshold %f\n",len,T);

			for (i=0; i<len; i++)
			{
				wsigma += SQR(data[i]);			/* sum of subband's squared wavelet coefficients */
			}
			wsigma /= len;						/* Now estimate of the subband's variance */

			wsigma -= SQR(sigma);					/* Now estimate of noise-free subband variance */
			if (wsigma > 0.0)						/* If <=0, use default */
				w = sigma*sigma / wsigma;			/* Bayes threshold */

			T = MIN(w,T);							/* Limit maximum to universal threshold */
			dp (2, "      BayesShrink threshold for %d data points: %f\n",len,T);

			return T;
			break;

		case 2:							/* SUREShrink */
			
			T = sigma * sqrt (2*SAFELOG2(len));		/* Upper bound is the universal threshold */
			p = calloc (len, sizeof (float));			/* Make a copy of the input data as it gets sorted */
			for (i=0; i<len; i++)
				p[i] = SQR(data[i]);					/* Squared wavelet coefficients */
			combsort_f (p, NULL, len);				/* Requires |wavelet coefficients| in ascending order */
			sumxx = 0;								/* Reset cumulative coefficients */

			for (i=0; i<len; i++)
			{
				sumxx += p[i];						/* Cumulative squared coefficients */
				w = sumxx + (len-i-1)*p[i];			/* Intermediate risk term */
				risk = (len - 2*i + w)/len;			/* Stein's Risk Estimator */
				dp (3, "      Coeff. %f, cumulative %f. Risk %f\n", sqrt(p[i]), w, risk);
				if (i==0)
				{
					minrisk=risk; j=0;
				}
				else if (minrisk > risk)
				{
					minrisk=risk; j=i;				/* Keep track of lowest risk, index j */
				}
			}

			w = sqrt(p[j]);							/* Return absolute value of coefficient with lowest risk */
			free (p);
			T = MIN (w,T);							/* Keep threshold below universal threshold */
			if (w==T)
				dp (2, "      SUREShrink threshold for %d data points: %f with risk %f\n",len,T, minrisk);
			else
				dp (2, "      SUREShrink threshold for %d data points: %f with risk %f limited to UT %f\n",len,w, minrisk, T);

			return T;
			break;

		case 3:									/* General Cross-Validation (GCV) */

			gcvmin=99999; gcvmint=0;
			max=data[0];
			for (i=0; i<len; i++)
			{
				max = MAX (max,fabs(data[i]));			/* Find maximum absolute value in data stream */
			}
			max /= 2;									/* We won't find the optimum at high threshold levels */

			for (w=0.0; w<=max; w+=max/200.0)				/* span threshold 0 to max in 200 steps. Good enough? */
			{
				sumxx = 0.0; cnt=0;
				for (i=0; i<len; i++)
				{
					buf = data[i];
					buf2 = wv_thresh (buf, w, 1); 			/* Always use hard thresholding in GCV computation */
					sumxx += (buf-buf2)*(buf-buf2);			/* Squared difference between thresholded and original */
					if (F_ISZERO(buf2)) cnt++;				/* Number of thresholded coefficients */
				}

				if (cnt>0)
					gcv = len*sumxx/cnt/cnt;
				else
					gcv = 0.08; 		/* 2 x (0.2)^2... why, exactly ?? */

				if (gcv < gcvmin) { gcvmin=gcv; gcvmint=w; }
				dp (3, "    GCV finder: Threshold %f, gcv=%f, coefficients zeroed: %d\n",w,gcv, cnt);

			}
			dp (2, "    --> GCV finder: Minimum GCV %f at threshold %f\n",gcvmin, gcvmint);
			return gcvmint;
	}
	return 0;
}






/**********************************************************************************

 ####    #####   #   #    #*#     #*#     #*#     #*#    #   #    #*#   
 #   #   #       #   #   #   #     #     #   #     #     #   #   #   #  
 #   #   #       ##  #   #   #     #     #         #     ##  #   #      
 #   #   ####    # # #   #   #     #      #*#      #     # # #   #  ##  
 #   #   #       #  ##   #   #     #         #     #     #  ##   #   #  
 #   #   #       #   #   #   #     #     #   #     #     #   #   #   #  
 ####    #####   #   #    #*#     #*#     #*#     #*#    #   #    #*#   

*/



/* This function performs denoising (= coefficient shrinking) on a 1D time series.
	It decomposes the signal into a "reasonable" number of dyadic levels and then
	applies a shrinking function to each level except the lowest lowpass subband.
	"reasonable" means that the lowest band will have 16 data points.

	wfunc is the mother wavelet function; see "set_subband" for how that works.
	wshrinkfunc is the threshold finder and shrinking function:
		0: Universal threshold, soft thresholding
		1: Universal threshold, hard thresholding
		2: Bayes shrink, soft thresholding
		3: Bayes shrink, hard thresholding
		4: SUREShrink, soft thresholding
		5: SUREShrink, hard thresholding
		6: GCV, soft thresholding
		7: GCV, hard thresholding
	tmult is a threshold multiplier for arbitrary proportional correction of the threshold

*/

void wvlt_1D_denoising  (float *data, int len, int wfunc, int wshrinkfunc, float tmult)
{
int M;							/* Needed to zero-pad */
float *paddata;
float T, sigma;
int i;
int xleft, xright, level, decomp_level;
WAVELET_TYPE wtype;
wavefilt wfilt;	



	M=2; decomp_level = -3;
	while (M<len)
	{
		M+=M;						/* Nearest power-of two */
		decomp_level++;				/* Keep track of reasonable decomposition levels */
	}

	if (decomp_level < 1) return;			/* Too small a dataset, don't do anything */

	paddata = calloc (M, sizeof(float));
	for (i=0; i<len; i++) paddata[i] = data[i];

	wtype = set_subband (wfunc);
	if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);

	dwt_1d (paddata, M, W_FORWARD, wfilt, (*subband));			/* Wavelet transform */

	/* Estimate noise component */

	sigma = estimate_sigma_1d (paddata, M);
	dp (2, "1D wavelet denoising: Estimated sigma %f\n", sigma);

	/* Begin dyadic level-by-level coefficient shrinking */

	xright = M; xleft = M/2;

	for (level=0; level<decomp_level; level++)
	{
		T = find_wsthreshold (&paddata[xleft], xleft, sigma, wshrinkfunc >> 1);

		for (i=xleft; i<xright; i++)
		{
			paddata[i] = wv_thresh (paddata[i], T*tmult, wshrinkfunc & 0x01);
		}

		xright >>=1; xleft >>= 1;
	}

	dwt_1d (paddata, M, W_INVERSE, wfilt, (*subband));			/* Inverse Wavelet transform */
	for (i=0; i<len; i++) data[i] = paddata[i];			/* Return data in-place */

}


/* A stand-alone function to obtain a noise estimate. This is a 1D function,
	but we can feed it 2D data, interpreted as 1D (this way, we need less padding)
	Moreover, we keep some coefficients fixed: We always use DAUB4 and one decomp_level
*/


float wvlt_noise_estimate  (float *data, int len)
{
int M;							/* Needed to zero-pad */
float *paddata;
float T, sigma;
int i;
int xleft, xright, level, decomp_level;
WAVELET_TYPE wtype;
wavefilt wfilt;	


	M=2; decomp_level = -3;
	while (M<len)
	{
		M+=M;						/* Nearest power-of two */
		decomp_level++;				/* Keep track of reasonable decomposition levels */
	}

	if (decomp_level < 1) return 0;			/* Too small a dataset, don't do anything */

	paddata = calloc (M, sizeof(float));
	for (i=0; i<len; i++) paddata[i] = data[i];

	wtype = set_subband (1);					/* THis sets DAUB4 */
	wfilt = wavelet_prepare (wtype);

	dwt_1d (paddata, M, W_FORWARD, wfilt, (*subband));			/* Wavelet transform */

	/* Estimate noise component */

	sigma = estimate_sigma_1d (paddata, M);
	dp (2, "1D wavelet noise detection: Estimated sigma %f\n", sigma);
	return sigma;
}



/* Wavelet shrinkage for image denoising.

	2D slice-by-slice wavelet shrinkage with different threshold finding algorithms.
	img must be a wavelet-transformed image with a decomposition given in decomp_level.
	For this reason, we assume the image type to be B_FLOAT (otherwise, we return doing nothing)
	(Note that this function performs the second step out of DWT -> shrinking -> IDWT)

	This function performs wavelet shrinkage according to wshrinkfunc and returns
	the shrunken coefficients for subsequent reconstruction.
	The function is
		0: Universal threshold, soft thresholding
		1: Universal threshold, hard thresholding
		2: Bayes shrink, soft thresholding
		3: Bayes shrink, hard thresholding
		4: SUREShrink, soft thresholding
		5: SUREShrink, hard thresholding
		6: GCV, soft thresholding
		7: GCV, hard thresholding


*/



void wavelet_shrink (image_type *img, int decomp_level, int wshrinkfunc, float tmult)
{
int x,y,z, xm,ym, x2,y2, i;
int level;
float *p;
float *tdata;
float sigma, T;			/* estimated noise, the threshold */

	if (img->imgtype!=B_FLOAT) return;
	p = img->data;

	for (z=0; z<img->zmax; z++)
	{
		sigma = estimate_sigma_2d (img, z);		/* Robust median estimator in HH (Donoho) for slice Z */

		xm=img->xmax; ym=img->ymax;		/* Full image size in top level */
		x2=xm; y2=ym;
		tdata = calloc (xm*ym, sizeof (float));		/* a very generous allocation ... */


		for (level=0; level<decomp_level; level++)
		{
			/* For each level, go over the H/L, H/H, and L/H areas separately.
				This also means having a separate threshold. */

			i=0;
			for (y=0; y<ym/2; y++)
				for (x=xm/2; x<xm; x++)
				{
					tdata[i++] = p[x + img->xmax*(y + img->ymax*z)];
				}
			T = find_wsthreshold (tdata, i, sigma, wshrinkfunc >> 1);
			for (y=0; y<ym/2; y++)
				for (x=xm/2; x<xm; x++)
				{
					p[x + img->xmax*(y + img->ymax*z)] = wv_thresh (p[x + img->xmax*(y + img->ymax*z)], T*tmult, wshrinkfunc & 0x01);
				}

			i=0;
			for (y=ym/2; y<ym; y++)
				for (x=xm/2; x<xm; x++)
				{
					tdata[i++] = p[x + img->xmax*(y + img->ymax*z)];
				}
			T = find_wsthreshold (tdata, i, sigma, wshrinkfunc >> 1);
			for (y=ym/2; y<ym; y++)
				for (x=xm/2; x<xm; x++)
				{
					p[x + img->xmax*(y + img->ymax*z)] = wv_thresh (p[x + img->xmax*(y + img->ymax*z)], T*tmult, wshrinkfunc & 0x01);
				}

			i=0;
			for (y=ym/2; y<ym; y++)
				for (x=0; x<xm/2; x++)
				{
					tdata[i++] = p[x + img->xmax*(y + img->ymax*z)];
				}
			T = find_wsthreshold (tdata, i, sigma, wshrinkfunc >> 1);
			for (y=ym/2; y<ym; y++)
				for (x=0; x<xm/2; x++)
				{
					p[x + img->xmax*(y + img->ymax*z)] = wv_thresh (p[x + img->xmax*(y + img->ymax*z)], T*tmult, wshrinkfunc & 0x01);
				}

			xm /=2; ym/=2;                          /* Shrink area by 2 for next level */

		}

		free (tdata);



	}

}




/**********************************************************/


